@@ -9,23 +9,26 @@
 
 The implementation is partially based on back-porting ``zipfile.Path`` (new in python 3.8)
 """
-from collections.abc import MutableMapping, Sequence
+from collections import abc
 from contextlib import contextmanager, suppress
 import io
 import itertools
 import os
 from pathlib import Path, PurePosixPath
 import posixpath
+import shutil
 import threading
 from types import TracebackType
 from typing import (
     IO,
     Any,
+    BinaryIO,
     Callable,
     Dict,
     Iterable,
     List,
     Optional,
+    Sequence,
     Set,
     Type,
     Union,
@@ -41,8 +44,12 @@ __all__ = (
     "FilteredZipInfo",
     "StopZipIndexRead",
     "read_file_in_zip",
+    "extract_file_in_zip",
+    "NOTSET",
 )
 
+NOTSET = ()
+
 
 class ZipPath:
     """A wrapper around ``zipfile.ZipFile``,
@@ -82,7 +89,9 @@ class ZipPath:
         at: str = "",  # pylint: disable=invalid-name
         allow_zip64: bool = True,
         compression: int = zipfile.ZIP_DEFLATED,
+        compresslevel: Optional[int] = None,
         name_to_info: Optional[Dict[str, zipfile.ZipInfo]] = None,
+        info_order: Sequence[str] = (),
     ):
         """Initialise a zip path item.
 
@@ -90,12 +99,20 @@ class ZipPath:
         :param at: the path within the zipfile (always use posixpath `/` separators)
         :param mode: the mode with which to open the zipfile,
             either read 'r', write 'w', exclusive create 'x', or append 'a'
+
+        write only options:
+
         :param allow_zip64: if True, the ZipFile will create files with ZIP64 extensions when needed
         :param compression: compression type
             ``zipfile.ZIP_STORED`` (no compression), ``zipfile.ZIP_DEFLATED`` (requires zlib),
             ``zipfile.ZIP_BZIP2`` (requires bz2) or ``zipfile.ZIP_LZMA`` (requires lzma)
         :param name_to_info: The dictionary for storing mappings of filename -> ``ZipInfo``,
-            if ``None``, defaults to ``{}``
+            if ``None``, defaults to ``{}``.
+            This can be used to implement on-disk storage of the zip central directory
+        :param info_order: ``ZipInfo`` for these file names will be written first
+            to the zip central directory.
+            These allows for faster reading of key files, in a zip that contains
+            many 1000s of files (see ``FilteredZipInfo``).
 
         """
         if posixpath.isabs(at):
@@ -114,15 +131,21 @@ class ZipPath:
                 path,
                 mode=mode,
                 compression=compression,
+                compresslevel=compresslevel,
                 allowZip64=allow_zip64,
                 name_to_info=name_to_info,
+                info_order=info_order,
             )
         else:
             self._filepath = path._filepath
             self._zipfile = path._zipfile
 
     def __str__(self):
-        return posixpath.join(self.root.filename, self.at)
+        return (
+            posixpath.join(self.root.filename, self.at)  # type: ignore[attr-defined]
+            if self.root.filename
+            else self.at
+        )
 
     def __repr__(self):
         return self.__repr.format(self=self)
@@ -237,12 +260,39 @@ class ZipPath:
         return self.__class__(self, at=posixpath.join(self.at, path))
 
     @contextmanager  # noqa: A003
-    def open(self, mode: str = "rb"):  # noqa: A003
-        """Open the file pointed by this path and return a file object."""
+    def open(  # noqa: A003
+        self, mode: str = "rb", *, compression=NOTSET, level=NOTSET, comment=NOTSET
+    ):
+        """Open the file pointed by this path and return a file object.
+
+        write only parameters:
+
+        :param compression: the ZIP compression method to use when writing the archive,
+                if not set use default value,
+                ZIP_STORED (no compression), ZIP_DEFLATED (requires zlib),
+                ZIP_BZIP2 (requires bz2) or ZIP_LZMA (requires lzma).
+        :param level: control the compression level to use when writing files to the archive
+                When using ZIP_DEFLATED integers 0 through 9 are accepted.
+                When using ZIP_BZIP2 integers 1 through 9 are accepted.
+        :param comment: A binary comment, stored in the central directory
+        """
         # zip file open misleading signals 'r', 'w', when actually they are byte mode
-        if mode not in {"rb", "wb"}:
+        zinfo: Union[str, zipfile.ZipInfo]
+        if mode == "wb":
+            zinfo = zipfile.ZipInfo(self.at)
+            zinfo.compress_type = (
+                self._zipfile.compression if compression is NOTSET else compression
+            )
+            zinfo._compresslevel = (  # type: ignore
+                self._zipfile.compresslevel if level is NOTSET else level
+            )
+            if comment is not NOTSET:
+                zinfo.comment = comment
+        elif mode == "rb":
+            zinfo = self.at
+        else:
             raise ValueError('open() requires mode "rb" or "wb"')
-        with self.root.open(self.at, mode=mode[0]) as handle:
+        with self.root.open(zinfo, mode=mode[0]) as handle:
             yield handle
 
     def _write(self, content: Union[str, bytes]):
@@ -439,14 +489,17 @@ class ZipPath:
             self._zipfile.extract(path=outpath, member=info)
 
 
-class FileList(Sequence):
+class FileList(abc.Sequence):
     """A list of ``zipfile.ZipInfo`` which mirrors the ``zipfile.ZipFile.NameToInfo`` mapping.
 
     For indexing, assumes that ``NameToInfo`` is an ordered dict.
     """
 
-    def __init__(self, name_to_info: Dict[str, zipfile.ZipInfo]):
+    def __init__(
+        self, name_to_info: Dict[str, zipfile.ZipInfo], info_order: Sequence[str] = ()
+    ):
         self._name_to_info = name_to_info
+        self._info_order = info_order
 
     def __getitem__(self, item):
         key = list(self._name_to_info)[item]
@@ -462,7 +515,12 @@ class FileList(Sequence):
         return key in self._name_to_info
 
     def __iter__(self):
-        for value in self._name_to_info.values():
+        for key in self._info_order:
+            if key in self._name_to_info:
+                yield self._name_to_info[key]
+        for key, value in self._name_to_info.items():
+            if key in self._info_order:
+                continue
             yield value
 
     def __reversed__(self):
@@ -481,23 +539,28 @@ class StopZipIndexRead(Exception):
     """An exception to signal that the reading of the index should be stopped."""
 
 
-class FilteredZipInfo(MutableMapping):
+class FilteredZipInfo(abc.MutableMapping):
     """A mapping which only stores pre-defined ``ZipInfo`` s.
 
     Once all required filenames are set, ``__setitem__`` will raise ``StopZipIndexRead``.
 
     """
 
-    def __init__(self, filenames: Set[str]):
+    def __init__(self, filenames: Set[str], max_infos: Optional[int] = None):
         self._dict: Dict[str, zipfile.ZipInfo] = {}
         self._filenames = set(filenames)
+        self._max_infos = max_infos
+        self._read = 0
 
     def __getitem__(self, name):
         return self._dict.__getitem__(name)
 
     def __setitem__(self, name, item):
+        self._read += 1
         if name in self._filenames:
             self._dict.__setitem__(name, item)
+        if self._max_infos and self._max_infos <= self._read:
+            raise StopZipIndexRead
         if set(self._dict) == self._filenames:
             raise StopZipIndexRead
 
@@ -541,20 +604,27 @@ class ZipFileExtra(zipfile.ZipFile):
         compresslevel: Optional[int] = None,
         *,
         strict_timestamps: bool = True,
-        name_to_info: Optional[MutableMapping] = None,
+        name_to_info: Optional[abc.MutableMapping] = None,
+        info_order: Sequence[str] = (),
     ):
         """Open the ZIP file with mode read 'r', write 'w', exclusive create 'x', or append 'a'.
 
         :param file: The zip file to use
         :param mode: The mode in which to open the zip file
         :param compression: the ZIP compression method to use when writing the archive
+                ZIP_STORED (no compression), ZIP_DEFLATED (requires zlib),
+                ZIP_BZIP2 (requires bz2) or ZIP_LZMA (requires lzma).
+        :param compresslevel: control the compression level to use when writing files to the archive
+                When using ZIP_DEFLATED integers 0 through 9 are accepted.
+                When using ZIP_BZIP2 integers 1 through 9 are accepted.
         :param allowZip64: If True, zipfile will create ZIP files that use the ZIP64 extensions,
             when the zipfile is larger than 4 GiB
-        :param compresslevel: control the compression level to use when writing files to the archive
         :param strict_timestamps: when set to False, allows to zip files older than 1980-01-01
 
         :param name_to_info: The dictionary for storing mappings of filename -> ``ZipInfo``,
             if ``None``, defaults to ``{}``
+        :param info_order: list of file names (if present) to write first to the central directory
+            Writing first means they can be identified faster
 
         """
         if mode not in ("r", "w", "x", "a"):
@@ -566,9 +636,11 @@ class ZipFileExtra(zipfile.ZipFile):
         self._didModify: bool = False
         self.debug: int = 0  # Level of printing: 0 through 3
         # Find file info given name
-        self.NameToInfo: Dict[str, zipfile.ZipInfo] = name_to_info or {}  # type: ignore
+        self.NameToInfo: Dict[str, zipfile.ZipInfo] = (
+            name_to_info if name_to_info is not None else {}  # type: ignore
+        )
         # List of ZipInfo instances for archive
-        self.filelist: List[zipfile.ZipInfo] = FileList(self.NameToInfo)  # type: ignore
+        self.filelist: List[zipfile.ZipInfo] = FileList(self.NameToInfo, info_order)  # type: ignore
         self.compression: int = compression  # Method of compression
         self.compresslevel: Optional[int] = compresslevel
         self.mode: str = mode
@@ -663,9 +735,13 @@ class ZipFileExtra(zipfile.ZipFile):
 
 
 def read_file_in_zip(
-    filepath: str, path: str, encoding: Optional[str] = "utf8"
+    filepath: str,
+    path: str,
+    encoding: Optional[str] = "utf8",
+    *,
+    search_limit: Optional[int] = None,
 ) -> Union[bytes, str]:
-    """Read a text based file from inside a zip file and return its content.
+    """Read a  file from inside a zip file and return its content.
 
     This function is optimised for cpu/memory performance,
     since it returns the file as soon as its index is found in the zip registry,
@@ -676,6 +752,7 @@ def read_file_in_zip(
     :param filepath: the path to the zip file
     :param path: the relative path within the zip file
     :param encoding: If not None, decode the bytes with this encoding
+    :param search_limit: Limit the search in the zip to the first n records
 
     :raises IOError: If the zip file cannot be read
     :raises FileNotFoundError: If the path in the zip file does not exist
@@ -683,7 +760,10 @@ def read_file_in_zip(
     """
     try:
         output = ZipFileExtra(
-            filepath, "r", allowZip64=True, name_to_info=FilteredZipInfo({path})
+            filepath,
+            "r",
+            allowZip64=True,
+            name_to_info=FilteredZipInfo({path}, max_infos=search_limit),
         ).read(path)
         if encoding is not None:
             return output.decode(encoding)
@@ -692,3 +772,45 @@ def read_file_in_zip(
         raise IOError(f"The input file cannot be read: {error}")
     except KeyError:
         raise FileNotFoundError(f"required file {path} is not included")
+
+
+def extract_file_in_zip(
+    filepath: str,
+    path: str,
+    handle: BinaryIO,
+    *,
+    buffer_size: Optional[int] = None,
+    search_limit: Optional[int] = None,
+) -> None:
+    """Extract file from inside a zip file and return its content.
+
+    This function is optimised for cpu/memory performance,
+    since it returns the file as soon as its index is found in the zip registry,
+    and does not construct the full index in memory.
+
+    For best performance, the path should have been stored near the start of the index.
+
+    :param filepath: the path to the zip file
+    :param path: the relative path within the zip file
+    :param handle: The handle to write to
+    :param buffer_size: Specify the byte chunk size for streaming
+    :param search_limit: Limit the search in the zip to the first n records
+
+    :raises IOError: If the zip file cannot be read
+    :raises FileNotFoundError: If the path in the zip file does not exist
+
+    """
+    try:
+        with ZipFileExtra(
+            filepath,
+            "r",
+            allowZip64=True,
+            name_to_info=FilteredZipInfo({path}, max_infos=search_limit),
+        ).open(path) as zip_handle:
+            shutil.copyfileobj(
+                zip_handle, handle, **({"length": buffer_size} if buffer_size else {})
+            )
+    except zipfile.BadZipfile as error:
+        raise IOError(f"The zip file cannot be read: {error}")
+    except KeyError:
+        raise FileNotFoundError(f"required file {path} is not included")
